#ifndef __MORPION_HPP__
#define __MORPION_HPP__

#include "game.hpp"
#include <random>
#include <array>
#include <iostream>
#include <memory>

namespace game
{
	struct morpion_state
	{
		uint16_t cross_bitboard = 0; //bitboard with the played moves of the cross
		uint16_t circle_bitboard = 0; //bitboard with the played moves of the circle
		uint8_t total_moves = 0; //Total played moves (<= 9)
		uint64_t possible_moves = 0x876543210; //List of possible moves left
		bool first_player_win = false; //First player is always the cross
		bool second_player_win = false; //Second player is always the circle
	};
	
	class morpion : public game<morpion_state>
	{
		public:
			morpion();
			morpion(const morpion& mor) = default;
			morpion& operator=(const morpion& mor) = default;
			bool end_of_game() const; //Is the game ended? (draw or won)
			int value(std::uint8_t player) const; //Returns if the player win, loose or nothing
			bool won(std::uint8_t player) const;
			bool lost(std::uint8_t player) const;
			bool draw(std::uint8_t player) const;
			uint8_t current_player() const; //The player that has to play next (at the beginning, the first player)
			std::uint16_t number_of_moves() const; //Number of moves that you can play
			std::uint16_t number_moves_played() const; //Number of moves played until now
			void play(std::uint16_t m); //Play the move m (updates the state). m is the position on the board of where you want to play
			void undo(std::uint16_t m) {}
			std::string player_to_string(std::uint8_t player) const; //String representation of a player
			std::string move_to_string(std::uint16_t m) const; //String representation of a move (for example, A1)
			std::string to_string() const; //String representation of the entire game
			std::set<int> to_input_vector() const;
			void from_input_vector(const std::set<int>& input);
			morpion_state get_state(); //Return the state
			void set_state(const morpion_state& state); //Replace the current state with the one passed as a parameter
			std::shared_ptr<game<morpion_state>> do_copy() const;
			std::uint64_t hash(std::uint16_t m) const;
			std::uint64_t hash() const;
			
		private:
			inline void update_win(); //Check if someone won and update the state
			inline bool has_won(uint16_t bitboard); //Check if the player whose bitboard was passed as a param has won
			inline bool get(uint16_t bitboard, uint8_t i, uint8_t j) const; //Get a case of the board
			inline void update_moves(); //Update the list of all possible moves
			
			const uint8_t CROSS = 0;
			const uint8_t CIRCLE = 1;

			morpion_state state;
			
			//WIN CONSTANTS
			const uint16_t ROW0_MASK = 504;
			const uint16_t ROW1_MASK = 455;
			const uint16_t ROW2_MASK = 63;
			const uint16_t COL0_MASK = 438;
			const uint16_t COL1_MASK = 365;
			const uint16_t COL2_MASK = 219;
			const uint16_t DIA0_MASK = 238;
			const uint16_t DIA1_MASK = 427;
			const uint16_t ALL_ONES = 511;

			static std::vector<std::vector<uint64_t>> cross_hash_values;
			static std::vector<std::vector<uint64_t>> circle_hash_values;
	};
	std::ostream& operator<<(std::ostream& os, const morpion& mor);
}

#endif
