#ifndef __MINMAX_HPP__
#define __MINMAX_HPP__

#include <chrono>
#include <vector>
#include "game.hpp"
#include "heuristic.hpp"
#include <iostream>
#include <limits>

namespace minmax
{
  struct cannot_prove : std::exception
  {
    const char* what() const noexcept { return "cannot prove the game theoritical value"; }
  };

  template <typename Game>
  class minmax
  {
  protected:
    Game& game;
    std::chrono::milliseconds milliseconds;
    const heuristic<Game>& h;
    std::vector<uint16_t> principal_variation;
    std::chrono::steady_clock::time_point start;

  public:
    minmax(Game& game, uint32_t milliseconds, const heuristic<Game>& h = null_heuristic<Game>())
      : game(game), milliseconds(milliseconds), h(h)
    {
    }

    int prove();
    double solve();
    uint16_t select_move();

  private:
    long prove(int depth);
    double solve(int depth);
  };

  template <typename Game>
  minmax<Game> make_minmax(Game& game, uint32_t milliseconds, const heuristic<Game>& h = null_heuristic<Game>())
  {
    return minmax<Game>(game, milliseconds, h);
  }
    
  template <typename Game>
  double minmax<Game>::solve()
  {
    return 0;
  }

  template <typename Game>
  double minmax<Game>::solve(int depth)
  {
    return 0;
  }

  template <typename Game>
  uint16_t minmax<Game>::select_move()
  {
    return 0;
  }

  const long UNKNOWN = std::numeric_limits<long>::max();
  
  template <typename Game>
  int minmax<Game>::prove()
  {
    start = std::chrono::steady_clock::now();
    uint16_t nb_moves = game.number_of_moves();
    long value[nb_moves];
    std::fill_n(value, nb_moves, UNKNOWN);
    int depth = 3;
    auto state = game.get_state();
    try
      {
	int nb_proved_moves = 0, nb_draws = 0;
	while (true) 
	  {
	    for (uint16_t move = 0; move < nb_moves; ++move) 
	      {
		if (value[move] != UNKNOWN) continue;
		game.play(move);
		value[move] = prove(depth);
		game.set_state(state);
		if (value[move] == -1) return 1;
		if (value[move] == 0) ++nb_draws;
		if (value[move] != UNKNOWN) ++nb_proved_moves;
	      }
	    if (nb_proved_moves == nb_moves)
	      {
		if (nb_draws != 0) return 0;
		return -1;
	      }
	    ++depth;
	  }
      }
    catch (cannot_prove fail)
      {
	game.set_state(state);
      }
    throw cannot_prove();
  }
  
  template <typename Game>
  long minmax<Game>::prove(int depth)
  {
    if (game.end_of_game()) return game.value(game.current_player());
    if (depth == 0) return UNKNOWN;
    if ((depth & 3) && std::chrono::steady_clock::now() >= start + milliseconds) throw cannot_prove();
    uint16_t nb_moves = game.number_of_moves();
    auto state = game.get_state();
    int nb_proved_moves = 0, nb_draws = 0;
    for (uint16_t move = 0; move < nb_moves; ++move)
      {
	game.play(move);
	long v = prove(depth - 1);
	game.set_state(state);
	if (v == -1) return 1;
	if (v == 0) ++nb_draws;
	if (v != UNKNOWN) ++nb_proved_moves;
      }
    if (nb_proved_moves == nb_moves)
      {
	if (nb_draws != 0) return 0;
	return -1;
      }
    return UNKNOWN;
  }
}

#endif
